
'===============================================================================
' Autor ............. Jrgen Schlottke, Schnaich-C.-Str. 46, D-W 2200 Elmshorn
' Zweck ............. Demonstration der Schachprogrammierung unter Turbo-Pascal
' Datum ....................... irgendwann 1991, als PD freigegeben am 18.01.93
' Version ................................................. ohne Versionsnummer
' FreeBasic Version ................................ Roland Chastain 18.01.2013
'===============================================================================

enum boolean
  true = (1=1)
  false = not true
end enum

dim shared as boolean quitting = false ' Soll das Spiel beendet werden
dim shared as integer halbzugzahl = 0  ' Anzahl der gemachten Zge
dim shared as integer rechentiefe = 2  ' Rechentiefe des Computers in Halbzgen
dim shared as integer maxtiefe    = 4  ' Maximale Rechentiefe bei Schlagzgen

const ohnewert     =   0 ' Wert eines leeren Felds
const bauernwert   =   2 ' Wert eines Bauernfelds
const laeuferwert  =   6 ' etc.
const springerwert =   7
const turmwert     =  10
const damenwert    =  19
const koenigswert  = 126
const ausserhalb   = 127 ' Wert eines Felds auerhalb des Brettes

' Bewegungsrichtungen der Figuren auf dem Spielfeld
dim shared as integer bauer   (1 to 4) = {+01,+02,-09,+11}
dim shared as integer springer(1 to 8) = {+12,+21,+19,+08,-12,-21,-19,-08}
dim shared as integer laeufer (1 to 4) = {+11,-11,+09,-09}
dim shared as integer turm    (1 to 4) = {-01,+01,-10,+10}
dim shared as integer koenig  (1 to 8) = {-01,+01,-10,+10,+11,-11,+09,-09}

rem 11 - springer(6) = -10
rem 88 + springer(2) = 109

type brt
  as integer brett(-10 to 109)      ' Spielfeld(1)
  as integer epfeld(-1 to 1)        ' En-Passant-Felder
  as integer zuglistevon(1 to 100)  ' Ausgangsfeld
  as integer zuglistenach(1 to 100) ' Endfeld
  as integer zugwert(1 to 100)      ' Bewertung
  as integer anz                    ' Anzahl der gefundenen Zge
  as integer material               ' Materialbilanz
  as integer farbe                  ' Farbe des am Zug befindlichen Spielers
end type

dim shared as brt spielbrett               ' Auf diesem Brett wird gespielt
dim shared as brt grundbrett               ' Ein Brett in Grundstellung
dim shared as integer wert, eom
dim shared as integer varianten            ' Anzahl der berechneten Zugvarianten
dim shared as integer mgzugvon(1 to 100)   ' Die bestmglichen Zge
dim shared as integer mgzugnach(1 to 100)
dim shared as integer mgzugwert(1 to 100)
dim shared as integer historyvon(1 to 700) ' Speichert 700 Halbzge
dim shared as integer historynach(1 to 700)

sub Vorspann
  dim as uByte f = FreeFile
  dim as string z
  if Open("vorspann.txt" for input as #f) then
    exit sub
  end if
  do until Eof(f)
    Line Input #f, z
    if Eof(f) then Print z; else Print z
  loop
  Sleep
  do
  loop until Inkey = ""
  Cls
end sub

function Stellungswert(byref b as brt, byval farbe as integer) as integer
' Zweck: Bewertung der aktuellen Brettstellung
  return b.material * farbe
end function ' Stellungswert

sub Grundstellung
' Zweck: Grundaufstellung der Figuren
  dim as integer i, j
  for i = -9 to 110
    grundbrett.brett(i) = ausserhalb
  next i
  for i = 1 to 8 : for j = 1 to 8
  	grundbrett.brett(10*i+j) = ohnewert
  next j : next i
  for i = 1 to 8
    grundbrett.brett(10*i+2) = bauernwert
  next i
  for i = 1 to 8
    grundbrett.brett(10*i+7) = -bauernwert
  next i
  grundbrett.brett(21) = +springerwert
  grundbrett.brett(71) = +springerwert
  grundbrett.brett(31) = +laeuferwert
  grundbrett.brett(61) = +laeuferwert
  grundbrett.brett(11) = +turmwert
  grundbrett.brett(81) = +turmwert
  grundbrett.brett(51) = +koenigswert
  grundbrett.brett(41) = +damenwert
  grundbrett.brett(28) = -springerwert
  grundbrett.brett(78) = -springerwert
  grundbrett.brett(38) = -laeuferwert
  grundbrett.brett(68) = -laeuferwert
  grundbrett.brett(18) = -turmwert
  grundbrett.brett(88) = -turmwert
  grundbrett.brett(58) = -koenigswert
  grundbrett.brett(48) = -damenwert
  grundbrett.material = 0
  grundbrett.farbe = 1
  spielbrett = grundbrett
end sub ' Grundstellung

sub Brettdarstellung(byref b as brt)
' Zweck: Darstellung des Bretts auf dem Monitor
  dim as string s = _
  "  A B C D E F G H " & Chr(13, 10) &                                         _
  "8                 " & Chr(13, 10) &                                         _
  "7                 " & Chr(13, 10) &                                         _
  "6                 " & Chr(13, 10) &                                         _
  "5                 " & Chr(13, 10) &                                         _
  "4                 " & Chr(13, 10) &                                         _
  "3                 " & Chr(13, 10) &                                         _
  "2                 " & Chr(13, 10) &                                         _
  "1                 " & Chr(13, 10)
  for y as integer = 8 to 1 step -1
    for x as integer = 1 to 8
      select case Abs(b.brett(10*x+y))
      	case ohnewert
      		s[20*(8-y)+2*x+20] = 176
      	case bauernwert
      		s[20*(8-y)+2*x+20] = Asc("B")
      	case springerwert
      		s[20*(8-y)+2*x+20] = Asc("S")
      	case laeuferwert
      		s[20*(8-y)+2*x+20] = Asc("L")
      	case turmwert
      		s[20*(8-y)+2*x+20] = Asc("T")
      	case damenwert
      		s[20*(8-y)+2*x+20] = Asc("D")
      	case koenigswert
      		s[20*(8-y)+2*x+20] = Asc("K")
      end select
      if b.brett(10*x+y) < 0 then
      	s[20*(8-y)+2*x+20] += 32
      end if
    next x
  next y
  Print
  Print s
  if b.farbe = 1 then
    Print "Wei";
  else
    Print "Schwarz";
  end if
  Print " ist am Zug. "
end sub

sub Setzen(byref b as brt, byval von as integer, byval nach as integer)
' Zweck: Einen Zug auf dem Spielbrett ausfhren
  if (Abs(nach - von) = 2) and (Abs(b.brett(von)) = bauernwert) then
  ' Wenn Bauer zwei Felder zieht, dann En-Passant ermglichen
    b.epfeld(b.farbe) = von + bauer(1) * b.farbe
  else
    b.epfeld(b.farbe) = 0
  end if
  if (nach = b.epfeld(-b.farbe)) and (b.brett(von) = bauernwert * b.farbe) then
  ' Wenn En-Passant-Zug, dann erst den Bauern des Gegners ein Feld zurcksetzen
     b.epfeld(b.farbe) = 0
     Setzen(b, von, (nach \ 10) * 10 + von mod 10)
     b.farbe = -b.farbe
     von = (nach \ 10) * 10 + von mod 10
  end if
  if ((von >= 51) and (von <= 58))                                             _
  and ((nach = 71)                                                             _
    or (nach = 31)                                                             _
    or (nach = 78)                                                             _
    or (nach = 38))                                                            _
  and (Abs(b.brett(von)) = koenigswert) then
  ' Beginnen mit Rochade
    b.brett(nach) = b.brett(von)
  ' Knig setzen
    b.brett(von) = 0
    if nach \ 10 = 7 then
    ' kurze Rochade
      von = (von mod 10) + 80
      nach = von - 20
    ' Turm geht 2 Felder
    else
    ' lange Rochade
      von = (von mod 10) + 10
      nach = von + 30
    ' Turm geht 3 Felder
    end if
    b.brett(nach) = b.brett(von)
    b.brett(von) = 0
  else
  ' Ein ganz normaler Zug
    b.material = b.material - b.brett(nach)
    b.brett(nach) = b.brett(von)
    b.brett(von) = 0
    if ((nach mod 10 = 1) or (nach mod 10 = 8))                                _
    and (Abs(b.brett(nach))) = bauernwert then
    ' Damenumwandlung
      b.brett(nach) = damenwert * b.farbe
      b.material = b.material + (damenwert - bauernwert) * b.farbe
    end if
  end if
  b.farbe = -b.farbe
end sub ' Setzen

sub Eintragen(byref b as brt, byval von as integer, byval nach as integer)
' Zweck: Einen gltigen Zug in die Zugliste eintragen
  if Abs(b.material) > 83 then exit sub
' Weiterspielen hat keinen Zweck mehr, Ende
  b.anz += 1
  b.zuglistevon(b.anz) = von
  b.zuglistenach(b.anz) = nach
end sub ' Eintragen

sub Zuggenerator_ohne(byref b as brt)
' Zweck: Alle mglichen Zgen ermitteln (ohne Rochade)
  dim as integer von, nach
  dim as integer i, j, k, l
  b.anz = 0 ' Anzahl der gefundenen Zge
  for j = 1 to 8 ' Alle Spielfelder durchlaufen
    for k = 1 to 8
      von = j * 10 + k ' von = 11 to 88
      if b.brett(von) <> ohnewert then
        select case b.brett(von) * b.farbe
        case bauernwert
          nach = von + bauer(1) * b.farbe
          if b.brett(nach) = ohnewert then
            eintragen(b, von, nach)
            if (b.farbe = 1) and (k = 2) or (b.farbe = -1) and (k = 7) then
              nach = von + bauer(2) * b.farbe
              if (b.brett(nach) = ohnewert) then
                Eintragen(b, von, nach)
              end if
            end if
          end if
          for i = 3 to 4
            nach = von + bauer(i) * b.farbe
            if ((-b.brett(nach) * b.farbe >= bauernwert)                       _
            and (-b.brett(nach) * b.farbe <= koenigswert))                     _
            or (nach = b.epfeld(-b.farbe)) then
              Eintragen(b, von, nach)
            end if
          next i
        case springerwert
          for i = 1 to 8
            nach = von + springer(i)
            if (-b.brett(nach) * b.farbe >= ohnewert)                          _
            and (-b.brett(nach) * b.farbe <= koenigswert) then
              Eintragen(b, von, nach)
            end if
          next i
        case laeuferwert
          for i = 1 to 4
            nach = von
            do
              nach += laeufer(i)
              if (-b.brett(nach) * b.farbe >= ohnewert)                        _
              and (-b.brett(nach) * b.farbe <= koenigswert) then
                Eintragen(b, von, nach)
              end if
            loop until b.brett(nach) <> ohnewert
          next i
        case turmwert
          for i = 1 to 4
            nach = von
            do
              nach += turm(i)
              if (-b.brett(nach) * b.farbe >= ohnewert)                        _
              and (-b.brett(nach) * b.farbe <= koenigswert) then
                Eintragen(b, von, nach)
              end if
            loop until b.brett(nach) <> ohnewert
          next i
        case koenigswert
          for i = 1 to 8
            nach = von + koenig(i)
            if (-b.brett(nach) * b.farbe >= ohneWert)                          _
            and (-b.brett(nach) * b.farbe <= Koenigswert) then
            ' Testen wo der gegnerische Knig steht
              l = 0
              do
                l += 1
              loop until (-b.brett(nach + koenig(l)) * b.farbe = koenigswert)  _
                      or (l = 8)
              if l = 8 then Eintragen(b, von, nach)
            end if
          next i
        case damenwert
          for i = 1 to 8
            nach = von
            do
              nach += koenig(i)
              if (-b.brett(nach) * b.farbe >= ohnewert)                        _
              and (-b.brett(nach) * b.farbe <= koenigswert) then
                Eintragen(b, von, nach)
              end if
            loop until b.brett(nach) <> ohnewert
          next i
        end select
      end if            
    next k
  next j
end sub ' Zuggenerator_ohne

function Schach(byref b as brt) as boolean
' Zweck: Testen, ob der Knig im Schach steht
' Aufruf nachdem gesetzt wurde
  b.farbe = -b.farbe
  zuggenerator_ohne(b)
  b.farbe = -b.farbe
  for i as integer = 1 to b.anz
    if Abs(b.brett(b.zuglistenach(i))) = koenigswert then
      return true ' wenn der Gegner den Knig erreicht
    end if
  next i
  return false
end function ' Schach

sub Rochade_testen(byref b as brt,                                             _
	                 byval vonkoenig as integer,                                 _
	                 byval schritt as integer,                                   _
	                 byval turmfeld as integer)
  dim brett1 as brt
  dim i as integer
  if b.brett(turmfeld) * b.farbe <> turmwert then exit sub
  i = vonkoenig + schritt
  do
    if b.brett(i) <> ohnewert then exit sub
    i += schritt
  loop until i = turmfeld
  brett1 = b
  if not Schach(brett1) then Eintragen(b, vonkoenig, vonkoenig + 2 * schritt)
' Steht der Knig auch nicht im Schach?
end sub

sub Zuggenerator(byref b as brt)
' Zweck: Ermitteln aller Zge incl. Rochade
' erst normale Zge ermitteln
  Zuggenerator_ohne(b)
' danach die Rochadezge
  if b.brett(51) * b.farbe = koenigswert then
    Rochade_testen(b, 51, 10, 81)
    Rochade_testen(b, 51,-10, 11)
  end if
  if b.brett(58) * b.farbe = koenigswert then
    Rochade_testen(b, 58, 10, 88)
    Rochade_testen(b, 58,-10, 18)
  end if
end sub ' Zuggenerator

function Maxwertung(byref b as brt, byval farbe as integer,                    _
                    byval tiefe as integer,                                    _
                    byval alfa as integer)                                     _
                    as integer
' Zweck: Rekursive Bewertung eines Spielbretts
  dim as integer minmax, wertung, zugnr
  dim as brt brett1
  dim as integer beta
  dim as boolean alphabetaabbruch
  
  Zuggenerator(b) ' alle Zge incl. Rochade
  zugnr = 0
  beta = -32000 * b.farbe * farbe
  alphabetaabbruch = false
  while (zugnr < b.anz) and not alphabetaabbruch
    zugnr += 1
    brett1 = b
    Setzen(brett1, b.zuglistevon(zugnr), b.zuglistenach(zugnr))

    if ((tiefe >= rechentiefe) and (b.brett(b.zuglistenach(zugnr)) = 0))       _
    or (tiefe = maxtiefe) then
      wertung = Stellungswert(brett1, farbe)
      varianten += 1
    else
      wertung = Maxwertung(brett1, farbe, tiefe+1, beta)
    end if
    
    if b.farbe = farbe then
      if wertung > beta then beta = wertung
      if beta > alfa then alphabetaabbruch = true
    else
      if wertung < beta then beta = wertung
      if beta < alfa then alphabetaabbruch = true
    end if
    
    b.zugwert(zugnr) = wertung
  wend
  Maxwertung = beta
end function ' Maxwertung

function Spielereingabe as boolean
' Zweck: Feststellen, was der Spieler will
  dim as integer i
  dim as brt brett1
  dim as string*4 eingabe
  dim as integer zug, code
  dim as integer von, nach
  start:
  code = 1
  while code <> 0
  	/'
	  Print "9999 = Ende"
	  Print "9900 = Computer zieht"
	  Print "9001 = Zurcknehmen"
	  '/
	  Input "Ihr Zug: ", eingabe
	  spielereingabe = true
	  
	  if eingabe[0] = Asc("9") then
	    spielereingabe = false
	    if eingabe = "9999" then quitting = true
	    if eingabe = "9900" then spielereingabe = true
	    if eingabe = "9001" then
	      'Cls
	      Grundstellung
	      if halbzugzahl > 2 then halbzugzahl -= 2 else halbzugzahl = 0
	      for i = 1 to halbzugzahl
	        Setzen(spielbrett, historyvon(i), historynach(i))
	      next i
	      Brettdarstellung(spielbrett)
	    end if
	    exit function
	  end if
	  
	  eingabe = UCase(eingabe)
	  if (InStr("ABCDEFGH", Chr(eingabe[0])) > 0)_
    and (InStr("ABCDEFGH", Chr(eingabe[2])) > 0) then
	    eingabe[0] -= 16
	    eingabe[2] -= 16
	  end if
	  zug = ValInt(eingabe)
	  if (zug >= 1111) and (zug <= 8888) then code = 0
  wend
  
	von = zug \ 100
	nach = zug mod 100
	Zuggenerator(spielbrett) ' Alle Zge incl. Rochade
	
	code = 1
	for i = 1 to spielbrett.anz
	  if zug = spielbrett.zuglistevon(i) * 100 + spielbrett.zuglistenach(i) then
      code = 0
    end if
	next i
	if (code = 1) then goto start
	
	brett1 = spielbrett
	Setzen(brett1, von, nach)
	brett1.farbe = -brett1.farbe
	if Schach(brett1) then goto start
	
	Setzen(spielbrett, von, nach)
	Brettdarstellung(spielbrett)
	halbzugzahl += 1
	historyvon(halbzugzahl) = von
	historynach(halbzugzahl) = nach
end function ' Spielereingabe

sub Zugausgeben(byval von as integer, byval nach as integer)
  dim zugstr as string*5
  dim zug as integer
	zug = von * 100 + nach
	zugstr = Str(zug)
	zugstr[0] += 16
	zugstr[2] += 16
	Print zugstr;
end sub

function feld_ist_in_grundstellung(byref b as brt,                             _
                                   byval feldnr as integer)                    _
                                   as boolean
  if b.brett(feldnr) = grundbrett.brett(feldnr) then
    return true
  else
    return false
  end if
end function

function Besterzug as integer
  dim brett1 as brt
  dim as integer i, j
  dim bestcounter as integer

  eom =0
  for i = 1 to spielbrett.anz
    if spielbrett.zugwert(i) = wert then
      eom += 1
      mgzugvon(eom) = spielbrett.zuglistevon(i)
      mgzugnach(eom) = spielbrett.zuglistenach(i)
      mgzugwert(eom) = 0
    end if
  next i

  bestcounter = 0
  besterzug = 1
  
  for i =1 to eom

    brett1 = spielbrett
    mgzugwert(i) = 0
    	
    if feld_ist_in_grundstellung(brett1, mgzugvon(i)) then
      mgzugwert(i) += 5
      if Abs(brett1.brett(mgzugvon(i))) = bauernwert then mgzugwert(i) += 2
    end if
    
    if (Abs(brett1.brett(mgzugvon(i))) = koenigswert) then mgzugwert(i) -= 10
    
    if (halbzugzahl < 32) and (                                                _
      (Abs(brett1.brett(mgzugvon(i))) = bauernwert) or                         _
      (Abs(brett1.brett(mgzugvon(i))) = laeuferwert) or                        _
      (Abs(brett1.brett(mgzugvon(i))) = springerwert)                          _
      ) then mgzugwert(i) += 20

    if (mgzugvon(i) \ 10 = 1)                                                  _
    or (mgzugvon(i) \ 10 = 8)                                                  _
    or (mgzugvon(i) mod 10 = 1)                                                _
    or (mgzugvon(i) mod 10 = 8) then mgzugwert(i) += 2

    if (mgzugnach(i) \ 10 = 1)                                                 _
    or (mgzugnach(i) \ 10 = 8)                                                 _
    or (mgzugnach(i) mod 10 = 1)                                               _
    or (mgzugnach(i) mod 10 = 8) then mgzugwert(i) -= 2
    
    brett1 = spielbrett
    Setzen(brett1, mgzugvon(i), mgzugnach(i))
    if feld_ist_in_grundstellung(brett1, mgzugnach(i)) then mgzugwert(i) -= 10

    brett1.farbe = -brett1.farbe
    Zuggenerator_ohne(brett1)
    
    for j = 1 to brett1.anz
      mgzugwert(i) += 1
      if brett1.zuglistenach(j) <> 0 then mgzugwert(i) += 1
    next j
    
    brett1.farbe = -brett1.farbe
    Zuggenerator_ohne(brett1)
    
    for j = 1 to brett1.anz
      mgzugwert(i) -= 1
      if brett1.zuglistenach(j) <> 0 then mgzugwert(i) -= 1
    next j

    if mgzugwert(i) >= bestcounter then
      bestcounter = mgzugwert(i)
      besterzug = i
    end if

  next i
end function ' Besterzug

sub Computerzug
' Zweck: Den besten Computerzug berechnen und ausfhren
	dim as integer i, j
	dim as boolean schon_im_schach
	
	Print("Bitte warten...")
	
	varianten = 0
  rechentiefe = 3 ' Rechentiefe des Computers in Halbzgen
  maxtiefe = 5    ' Maximale Rechentiefe bei Schlagzgen
	
	wert = Maxwertung(spielbrett, spielbrett.farbe, 1, 32000)
  i = Besterzug
  
  /'
	Print("Berechnete Zge: ");
	for j = 1 to eom
	  if j < 33 then
	  	Zugausgeben(mgzugvon(j), mgzugnach(j))
	  else
	    if j = 33 then Print "etc."
	  end if
	next j 
	'/
  
	halbzugzahl += 1
	historyvon(halbzugzahl) = mgzugvon(i)
	historynach(halbzugzahl) = mgzugnach(i)
  
	schon_im_schach = Schach(spielbrett)
	Setzen(spielbrett, mgzugvon(i), mgzugnach(i))
	spielbrett.farbe = -spielbrett.farbe
	
	if Schach(spielbrett) then
	  if schon_im_schach then
	    Print "Ich bin SCHACHMATT"
	  else
	    Print "Es ist ein PATT"
	  end if
	  Sleep
	  'end
	  exit sub
	end if
	
	Print("Mein Zug: ");
	Zugausgeben(mgzugvon(i), mgzugnach(i))
	Print
	spielbrett.farbe = -spielbrett.farbe
	if Schach(spielbrett) then Print "SCHACH!"
	Brettdarstellung(spielbrett)
end sub ' Computerzug

'===============================================================================

rem BEGINN DES HAUPTPROGRAMMS: SCHACHSPIEL

Vorspann
Grundstellung
Brettdarstellung(spielbrett)
do
  if Spielereingabe then Computerzug
  Sleep 100
loop until quitting

rem ENDE DES HAUPTPROGRAMMS: SCHACHSPIEL
